""" Run a jfx_bridge_ida server for external python environments to interact with """

import functools
import inspect
import logging
import os
import subprocess
import types

from jfx_bridge import bridge
from jfx_bridge_ida_port import DEFAULT_SERVER_PORT

import idaapi

# pull in idc and idautils so they're available for remote_evals without extra work
import idc
import idautils

# try to do the same thing with sark, but it's fine if it's not present
try:
    import sark
except ImportError:
    pass

# IDA specific hooking
# IDA 7.2 requires all APIs not explicitly marked THREAD_SAFE to be called from the main thread. It provides
# the idaapi.execute_sync function to carry this out, which takes a python callable to run, and returns an
# int.
#
# To handle this, we hook the local_call/local_eval functions in the bridge, and wrap whatever we were going to call in
# a bound callable that provides a way to return an arbitrary result, then pass that bound callable to execute_sync.


class WrapperReturn(object):
    """ class to return an arbitrary result """

    result = None
    exception = None  # if not None, we hit an exception, and result isn't valid


def wrapper_execute_on_main_thread(prepped_function, return_object):
    """ Wrapper to handle calling the prepped function (with bound arguments) on the main IDA thread after being run through execute_sync.
        Will execute the prepped function and pass the result (or any exception) back to the return_object provided """
    try:
        return_object.result = prepped_function()
    except Exception as e:
        return_object.exception = e

    return 0


def call_execute_sync_and_get_result(
    prepped_function, execute_sync_flag=idaapi.MFF_WRITE
):
    """ we use MFF_WRITE to make sure that regardless of whether the operation reads/writes/ignores the database, it'll be fine """
    # if the caller has deliberately specified no execute_sync flag, don't call it - just call the prepped function directly and return the result
    if execute_sync_flag is None:
        return prepped_function()

    # where we get our result
    return_object = WrapperReturn()

    # bind the prepped function bound with args into the callable wrapper we'll pass to execute sync, along with the return object
    prepped_wrapper = functools.partial(
        wrapper_execute_on_main_thread, prepped_function, return_object
    )

    # run it on the main thread
    idaapi.execute_sync(prepped_wrapper, execute_sync_flag)

    if return_object.exception is not None:
        # something went wrong. reraise the exception on this thread, so it'll get passed back
        raise return_object.exception

    # and we're done!
    return return_object.result


# we don't wrap these, because we know they're safe and/or being wrapped may affect their operation
NO_WRAP_LIST = [idaapi.execute_sync, idaapi.is_main_thread]


def is_ida_module(module_name):
    """ Return true if it's an IDA-related module we want to execute_sync callables from
        sark, idaapi, idautils, idc - any ida_ (e.g., ida_kernwin), _ida_*.pyd, and a few edge cases (idadex.py, idc_bc695.py)
    """
    return module_name.startswith(("_ida_", "ida", "idc")) or module_name == "sark"


def should_execute_sync(target_callable):
    """ Return true if target_callable is IDA-related and is something we should use execute_sync to run on the main thread """

    if (
        target_callable not in NO_WRAP_LIST
    ):  # exclude the ones we know we don't want to wrap
        module = inspect.getmodule(
            target_callable
        )  # easy mode - does it have a module inspect can tell us about
        if module is None:
            if isinstance(target_callable, functools.partial):
                # partials - get the module from the function that's been wrapped
                module = inspect.getmodule(target_callable.func)
            elif "method-wrapper" in str(type(target_callable)):
                # bound to an instance of an object (in __self__)
                obj = target_callable.__self__
                if isinstance(obj, types.GeneratorType):
                    # generators need some additional inspection to see if they're IDA code
                    module = os.path.basename(obj.gi_code.co_filename).split(".")[0]
                else:  # expect the object to have __module__
                    module = inspect.getmodule(target_callable)
            elif "wrapper_descriptor" in str(type(target_callable)):
                # slot wrappers for things like base object implementation of __str__ - has a class in __objclass__
                module = inspect.getmodule(target_callable.__objclass__)

        if module is None:
            if isinstance(target_callable, type):
                if hasattr(target_callable, "__module__"):
                    if isinstance(
                        target_callable.__module__, bridge.STRING_TYPES
                    ):  # if python2: unicode as well as str
                        # this is a type created for a remote bridge (e.g., inheriting from a local type)
                        # check to see if any of the bases are IDA related
                        ida_parent = False
                        for base in target_callable.__bases__:
                            base_module = inspect.getmodule(base)
                            if base_module is not None and is_ida_module(
                                base_module.__name__
                            ):
                                ida_parent = True
                                break

                        # if they are, check to see if __init__ is overriden by a bridged function or not - if it isn't, instantiating the type could execute IDA code and we should execute_sync
                        if ida_parent:
                            if (
                                not "__init__" in target_callable.__dict__
                                or not bridge._is_bridged_object(
                                    target_callable.__dict__["__init__"]
                                )
                            ):
                                return True

                        # not IDA-related base type, or has a bridged __init__ (if that calls an ida function, we can execute_sync that when it comes in)
                        return False

            # don't know what this is, complain
            raise Exception(
                "Unknown module for : "
                + str(target_callable)
                + " "
                + str(type(target_callable))
                + " "
                + str(target_callable.__dict__)
            )

        if inspect.ismodule(module):
            module = module.__name__

        return is_ida_module(module)

    return False


def hook_local_call(bridge_conn, target_callable, *args, **kwargs):
    """ Hook the real local_call and see if we should use execute_sync to run the target on the main thread """

    execute_sync_flag = None  # if it's not IDA-related, we'll skip the execute_sync
    if should_execute_sync(target_callable):
        execute_sync_flag = idaapi.MFF_WRITE

    # Possible future - pull an execute sync flag kwarg from the callable args, to allow explicitly marking a particular call as MFF_READ or not needing execute sync

    # bind the target to the arguments
    prepped_function = functools.partial(target_callable, *args, **kwargs)

    return call_execute_sync_and_get_result(
        prepped_function, execute_sync_flag=execute_sync_flag
    )


def hook_local_eval(bridge_conn, eval_expr, eval_globals, eval_locals):
    """ Hook the real local_eval and use execute_sync to run the eval on the main thread """

    # first, bind the eval function to the arguments
    prepped_function = functools.partial(eval, eval_expr, eval_globals, eval_locals)

    return call_execute_sync_and_get_result(prepped_function)


def run_server(
    server_host=bridge.DEFAULT_HOST,
    server_port=DEFAULT_SERVER_PORT,
    response_timeout=bridge.DEFAULT_RESPONSE_TIMEOUT,
    background=True,
):
    """ Run a jfx_bridge_ida server (forever)
        server_host - what address the server should listen on
        server_port - what port the server should listen on
        response_timeout - default timeout in seconds before a response is treated as "failed"
        background - false to run the server in this thread (will lock up GUI), true for a new thread
    """
    server = bridge.BridgeServer(
        server_host=server_host,
        server_port=server_port,
        loglevel=logging.INFO,
        response_timeout=response_timeout,
        local_call_hook=hook_local_call,
        local_eval_hook=hook_local_eval,
    )

    if background:
        server.start()
        print(
            "Server launching in background - will continue to run after launch script finishes...\n"
        )
    else:
        server.run()


def run_script_across_bridge(script_file, python="python", argstring=""):
    """ Spin up a jfx_bridge_ida_Server and spawn the script in external python to connect back to it. Useful in scripts being triggered from
        inside an older IDA that's stuck in python2 or 32-bit python.

        The called script needs to handle the --connect_to_host and --connect_to_port command-line arguments and use them to start
        a jfx_bridge_ida client to talk back to the server.

        Specify python to control what the script gets run with. Defaults to whatever python is in the shell - if changing, specify a path
        or name the shell can find.
        Specify argstring to pass further arguments to the script when it starts up.
    """

    # spawn a jfx_bridge_ida server - use server port 0 to pick a random port
    server = bridge.BridgeServer(
        server_host="127.0.0.1",
        server_port=0,
        loglevel=logging.INFO,
        local_call_hook=hook_local_call,
        local_eval_hook=hook_local_eval,
    )
    # start it running in a background thread
    server.start()

    try:
        # work out where we're running the server
        server_host, server_port = server.get_server_info()

        print("Running " + script_file)

        # spawn an external python process to run against it
        try:
            output = subprocess.check_output(
                "{python} {script} --connect_to_host={host} --connect_to_port={port} {argstring}".format(
                    python=python,
                    script=script_file,
                    host=server_host,
                    port=server_port,
                    argstring=argstring,
                ),
                stderr=subprocess.STDOUT,
                shell=True,
            )
            print(output)
        except subprocess.CalledProcessError as exc:
            print("Failed ({}):{}".format(exc.returncode, exc.output))

        print(script_file + " completed")

    finally:
        # when we're done with the script, shut down the server
        server.shutdown()


if __name__ == "__main__":
    run_server(background=True)



""" Shutdown a running jfx_bridge_ida server cleanly """

from jfx_bridge import bridge
from jfx_bridge_ida_port import DEFAULT_SERVER_PORT

if __name__ == "__main__":
    print("Requesting server shutdown")
    client = bridge.BridgeClient(
        connect_to_host="127.0.0.1", connect_to_port=DEFAULT_SERVER_PORT
    )

    print(client.remote_shutdown())



""" Handle installing the jfx_bridge_ida server scripts to a specified directory.

    Note: we don't install the jfx_bridge module itself - pip install that into the python used by your IDA manually
"""
import argparse
import os
import pkg_resources

JFX_BRIDGE = "jfx_bridge"
IDA_BRIDGE = "jfx_bridge_ida"
SERVER_DIR = "server"


def do_install(install_dir):
    if not os.path.isdir(install_dir):
        os.makedirs(install_dir)

    # list the files from jfx_bridge_ida server directory
    server_files = [
        f
        for f in pkg_resources.resource_listdir(IDA_BRIDGE, SERVER_DIR)
        if f not in ["__init__.py", "__pycache__"]
    ]

    print("Installing jfx_bridge_ida server scripts...")

    # write out the jfx_bridge_ida server files directly in the install dir
    for f in server_files:
        dest_path = os.path.join(install_dir, f)
        with pkg_resources.resource_stream(
            IDA_BRIDGE, SERVER_DIR + "/" + f
        ) as resource:
            with open(dest_path, "wb") as dest:
                print("\t" + dest_path)
                dest.write(resource.read())

    print("Install completed")
    print(
        "!! If your IDA is using a different python instance to this one, remember to pip install jfx-bridge in that python instance !!"
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Install jfx_bridge_ida server scripts"
    )
    parser.add_argument(
        "install_dir",
        help="A directory where you want to store scripts to run from IDA",
    )

    args = parser.parse_args()

    do_install(args.install_dir)



import unittest
import os
import logging
import functools

import jfx_bridge
import jfx_bridge_ida
from jfx_bridge_ida.server import jfx_bridge_ida_port


def print_stats(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        start_stats = self.test_bridge.get_stats()
        func(self, *args, **kwargs)
        print(
            "\n{}:\n\t{}\n".format(
                func.__name__, self.test_bridge.get_stats() - start_stats
            )
        )

    return wrapper
class TestIDABridge(unittest.TestCase):
    """ Assumes there's an IDA bridge server running at DEFAULT_SERVER_PORT """

    @classmethod
    def setUpClass(cls):
        port = int(os.environ.get("TEST_PORT", jfx_bridge_ida_port.DEFAULT_SERVER_PORT))
        cls.test_bridge = jfx_bridge_ida.IDABridge(
            connect_to_port=port, loglevel=logging.DEBUG, record_stats=True
        )

    @print_stats
    def test_get_screen_ea(self):
        """ Confirm we can get the current address - this is not a threadsafe function, so if this works, our execute_sync wrapper is working """
        idaapi = self.test_bridge.get_idaapi()
        idaapi.get_screen_ea()

    @print_stats
    def test_threadsafe_generator(self):
        """ Confirm we can wrap a generator to use execute_sync """
        idautils = self.test_bridge.get_idautils()
        next(idautils.Segments())

    @print_stats
    def test_remote_eval(self):
        """ Confirm that we can do a remote_eval fine - that the ida functions are still wrapped """
        idc = self.test_bridge.get_idc()
        idautils = self.test_bridge.get_idautils()

        # pre IDA 7 compat
        get_func_name_fn_name = (
            "get_func_name" if "get_func_name" in dir(idc) else "GetFunctionName"
        )
        segment_end_fn_name = "get_segm_end" if "get_segm_end" in dir(idc) else "SegEnd"
        for segea in idautils.Segments():
            self.test_bridge.bridge.remote_eval(
                "[(funcea, idc.{}(funcea)) for funcea in idautils.Functions(segea, idc.{}(segea))]".format(
                    get_func_name_fn_name, segment_end_fn_name
                ),
                segea=segea,
                timeout_override=20,
            )

    @print_stats
    def test_remote_eval_exception(self):
        """ Confirm that we can still get exceptions from the remote_eval after execute_sync """
        with self.assertRaises(jfx_bridge.bridge.BridgeException):
            self.test_bridge.bridge.remote_eval("1/0")

    @print_stats
    def test_inheritance(self):
        """ Check we can inherit from an IDA object and override a method (e.g., __init__) which calls another IDA function (e.g., the parent __init__)
            without getting caught in an execute_sync deadlock """
        idaapi = self.test_bridge.get_idaapi()

        class TestListener(idaapi.UI_Hooks):
            called = False

            def __init__(self):
                idaapi.UI_Hooks.__init__(self)
                self.called = True

        t = TestListener()

        self.assertTrue(t.called)

    @print_stats
    def test_callback(self):
        """ Test we can get a callback 
        
        Expected failure on pre 7.2 - screen_ea_changed doesn't exist
        """
        idaapi = self.test_bridge.get_idaapi()

        class TestListener(idaapi.UI_Hooks):
            called = False

            def screen_ea_changed(self, new_ea, prev_ea):
                self.called = True
                return 0

        t = TestListener()
        try:
            t.hook()
            idaapi.jumpto(idaapi.get_screen_ea() + 0x10)
        finally:
            t.unhook()

        self.assertTrue(t.called)

    @print_stats
    def test_function_help(self):
        """ Check that we can view the function help """
        idaapi = self.test_bridge.get_idaapi()
        self.assertTrue("get_screen_ea()" in idaapi.get_screen_ea.__doc__)

    @print_stats
    def test_type_instantiate(self):
        """ check we can instantiate an IDA type that isn't threadsafe in its init 
        
        Expected failure in IDA pre 7 - no idadex. Haven't found another class with a non-threadsafe init
        """
        idadex = self.test_bridge.remote_import("idadex")
        try:
            idadex.Dex()
        except Exception as e:  # idadex.Dex will except if not operating on a dex - only care about the main thread error
            self.assertTrue(
                "Function can be called from the main thread only" not in str(e)
            )

    @print_stats
    def test_type_instantiate_inherited_but_non_overridden_init(self):
        """ check we can instantiate an type we've created from an IDA type that isn't threadsafe in its init, and we haven't overridden that init 
        
        Expected failure in IDA pre 7 - no idadex. Haven't found another class with a non-threadsafe init
        """
        idadex = self.test_bridge.remote_import("idadex")

        class TestDex(idadex.Dex):
            pass

        try:
            t = TestDex()
        except Exception as e:  # idadex.Dex will except if not operating on a dex - only care about the main thread error
            self.assertTrue(
                "Function can be called from the main thread only" not in str(e),
                "Original __init__ not running on main thread",
            )

    @print_stats
    def test_type_instantiate_inherited_with_overridden_init(self):
        """ check we can instantiate an type we've created from an IDA type that isn't threadsafe in its init, we've overridden that init, but we're still calling it in our init 
        
        Expected failure in IDA pre 7 - no idadex. Haven't found another class with a non-threadsafe init
        """
        idadex = self.test_bridge.remote_import("idadex")

        parent = self

        class TestDex(idadex.Dex):
            def __init__(self):
                try:
                    idadex.Dex.__init__(self)
                except Exception as e:  # idadex.Dex will except if not operating on a dex - only care about the main thread error
                    parent.assertTrue(
                        "Function can be called from the main thread only"
                        not in str(e),
                        "Original __init__ not running on main thread",
                    )

        t = TestDex()



""" Run a jfx_bridge_ida client to remotely communicate with a jfx_bridge_ida server
running in an IDA instance """

import sys

from jfx_bridge import bridge

from .server.jfx_bridge_ida_port import DEFAULT_SERVER_PORT


class IDABridge(bridge.BridgeClient):
    idc = None
    idaapi = None
    idautils = None
    sark = None

    def __init__(
        self,
        connect_to_host=bridge.DEFAULT_HOST,
        connect_to_port=DEFAULT_SERVER_PORT,
        loglevel=None,
        response_timeout=bridge.DEFAULT_RESPONSE_TIMEOUT,
        do_import=True,
        hook_import=False,
        record_stats=False,
        **kwargs
    ):
        """ Set up a bridge. Default settings connect to the default jfx_bridge_ida server

        loglevel for what logging messages you want to capture

        response_timeout is how long to wait for a response before throwing an exception, in seconds
        
        If do_import is true, gets the remote idaapi, idc and idautils modules and loads them into sys.modules,
        to make importing sark locally easy
        
        Set hook_import to True to use the jfx_bridge import hooking feature and allow easy importing. This may cause issues
        with running multiple IDABridges in the same process.
        
        Set record_stats to True to record statistics about the bridge to help improve performance. Get the stats with the get_stats() function
        """
        super().__init__(
            connect_to_host=connect_to_host,
            connect_to_port=connect_to_port,
            loglevel=loglevel,
            response_timeout=response_timeout,
            hook_import=hook_import,
            record_stats=record_stats,
            **kwargs
        )

        if do_import:
            self.get_idaapi(do_import=True)
            self.get_idc(do_import=True)
            self.get_idautils(do_import=True)

    def get_idaapi(self, do_import=False):
        """ Get the idaapi from the remote connection. 
        If do_import is true, store it into sys.modules so other things can import it easily
         - you probably don't want this, use do_import when setting up the bridge instead """
        if self.idaapi is None:
            self.idaapi = self.remote_import("idaapi")

        if do_import:
            sys.modules["idaapi"] = self.idaapi

        return self.idaapi

    def get_idc(self, do_import=False):
        """ Get the idc module from the remote connection. 
        If do_import is true, store it into sys.modules so other things can import it easily
        - you probably don't want this, use do_import when setting up the bridge instead """
        if self.idc is None:
            self.idc = self.remote_import("idc")

        if do_import:
            sys.modules["idc"] = self.idc

        return self.idc

    def get_idautils(self, do_import=False):
        """ Get the idautils module from the remote connection. 
        If do_import is true, store it into sys.modules so other things can import it easily
        - you probably don't want this, use do_import when setting up the bridge instead """
        if self.idautils is None:
            self.idautils = self.remote_import("idautils")

        if do_import:
            sys.modules["idautils"] = self.idautils

        return self.idautils

    def get_sark(self, do_import=False):
        """ Import sark from the remote IDA context 
            If do_import is true, store it into sys.modules so other things can import it easily - you probably don't want this """
        if self.sark is None:
            self.sark = self.remote_import("sark")

        if do_import:
            sys.modules["sark"] = self.sark

        return self.sark



import setuptools
import subprocess

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jfx_bridge_ida",
    version=subprocess.check_output("git describe", shell=True).decode("utf-8").strip(),
    author="justfoxing",
    author_email="justfoxingprojects@gmail.com",
    description="RPC bridge from Python to IDA Python interpreter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/justfoxing/jfx_bridge_ida",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["jfx_bridge>0.4.2"],
)
